/*
 * http_mime.c: Sends/gets MIME headers for requests
 * 
 * Rob McCool
 * 
 */


#include "httpd.h"

struct mime_ext {
    char *ext;
    char *ct;
    struct mime_ext *next;
};

/* Hash table */
struct mime_ext *types[26];
struct mime_ext *forced_types;

int content_length;
char content_type[MAX_STRING_LEN];

char location[MAX_STRING_LEN];
static char last_modified[MAX_STRING_LEN];

char auth_line[MAX_STRING_LEN];

void hash_insert(struct mime_ext *me) {
    register int i = me->ext[0] - 'a';
    register struct mime_ext *p, *q;

    if(!(q=types[i])) {
        types[i]=me;
        return;
    }
    if((!(p=q->next)) && (strcmp(q->ext,me->ext) >= 0)) {
        types[i]=me;
        me->next=q;
        return;
    }
    while(p) {
        if(strcmp(p->ext,me->ext) >= 0) break;
        q=p;
        p=p->next;
    }
    me->next=p;
    q->next=me;
}

void kill_mime() {
    register struct mime_ext *p,*q;
    register int x;

    for(x=0;x<26;x++) {
        p=types[x];
        while(p) {
            free(p->ext);
            free(p->ct);
            q=p;
            p=p->next;
            free(q);
        }
    }
}

void init_mime() {
    char l[MAX_STRING_LEN],w[MAX_STRING_LEN],*ct;
    FILE *f;
    register struct mime_ext *me;
    register int x,y;

    if(!(f = fopen(types_confname,"r"))) {
        fprintf(stderr,"httpd: could not open mime types file %s\n",
                types_confname);
        perror("fopen");
        exit(1);
    }

    for(x=0;x<26;x++) 
        types[x] = NULL;
    forced_types = NULL;

    while(!(getline(l,MAX_STRING_LEN,f,0))) {
        if(l[0] == '#') continue;
        getword(w,l,' ');
        if(l[0] == '\0') continue;
        ct = (char *)malloc(sizeof(char) * (strlen(w) + 1));
        strcpy(ct,w);

        while(l[0]) {
            getword(w,l,' ');
            if(w[0] && (w[0] != ' ')) {
                me = (struct mime_ext *)malloc(sizeof(struct mime_ext));
                me->ext = (char *)malloc(sizeof(char) * (strlen(w)+1));
                for(x=0;w[x];x++)
                    me->ext[x] = (islower(w[x]) ? w[x] : tolower(w[x]));
                me->ext[x] = '\0';
                me->ct=strdup(ct);
                me->next=NULL;
                hash_insert(me);
            }
        }
        free(ct);
    }

    fclose(f);
}

void dump_types() {
    struct mime_ext *p;
    register int x;

    for(x=0;x<26;x++) {
        p=types[x];
        while(p) {
            printf("ext %s: %s\n",p->ext,p->ct);
            p=p->next;
        }
    }
    p=forced_types;
    while(p) {
        printf("file %s: %s\n",p->ext,p->ct);
        p=p->next;
    }
}

int is_content_type(char *type) {
    return(!strcmp(content_type,type));
}

void set_content_type(char *fn) {
    int i;
    struct mime_ext *p;

    p=forced_types;

    while(p) {
        if(!strcmp(p->ext,fn)) {
            strcpy(content_type,p->ct);
            return;
        }
        p=p->next;
    }

    if((i=rind(fn,'.')) < 0) {
        strcpy(content_type,default_type);
        return;
    }

    p=types[(islower(fn[++i]) ? fn[i] : tolower(fn[i])) - 'a'];
    while(p) {
        if(!strcasecmp(&fn[i],p->ext)) {
            strcpy(content_type,p->ct);
            return;
        }
        p=p->next;
    }
    strcpy(content_type,default_type);
}

int scan_content_type(FILE *f, FILE *fd) {
    char l[MAX_STRING_LEN];
    char w[MAX_STRING_LEN];

    while(1) {
        if(getline(l,MAX_STRING_LEN,f,0))
            die(SERVER_ERROR,"httpd: script failed to produce output",fd);

        if(l[0] == '\0') return is_url(location);
        getword(w,l,':');
        if(!strcasecmp(w,"Content-type"))
            sscanf(l,"%s",content_type);
        else if(!strcasecmp(w,"Location")) {
            getword(w,l,'\0');
            sscanf(w,"%s",location);
        }
    }
}

void add_type(char *fn, char *t) {
    struct mime_ext *n=(struct mime_ext *)malloc(sizeof(struct mime_ext));

    n->ext = strdup(fn);
    n->ct = strdup(t);
    n->next = forced_types;
    forced_types = n;
}

void set_content_length(int l) {
    content_length = l;
}

void set_last_modified(time_t t) {
    struct tm *tt;
    char *ts;

    tt = gmtime(&t);
    ts = asctime(tt);
    ts[strlen(ts) - 1] = '\0';
    strcpy(last_modified,ts);
}

void get_mime_headers(FILE *fd) {
    char l[MAX_STRING_LEN];
    char w[MAX_STRING_LEN];

    content_type[0] = '\0';
    last_modified[0] = '\0';
    content_length = -1;
    auth_line[0] = '\0';

    while(1) {
        getline(l,MAX_STRING_LEN,fd,timeout);
        getword(w,l,':');
        if(!strcasecmp(w,"Content-type")) 
            sscanf(l,"%s",content_type);
        if(!strcasecmp(w,"Authorization"))
            strcpy(auth_line,l);
        if(!strcasecmp(w,"Content-length"))
            sscanf(l,"%d",&content_length);
        else if(!l[0])
            return;
    }
}

void dump_default_header(FILE *fd) {
    fprintf(fd,"Date: %s GMT%c",get_gmt(),LF);
    fprintf(fd,"Server: %s%c",SERVER_VERSION,LF);
    fprintf(fd,"MIME-version: 1.0%c",LF);
}

void send_mime_headers(FILE *fd) {
    fprintf(fd,"%s %s%c",SERVER_PROTOCOL,
            (location[0] ? "302 Found" : "200 OK"),LF);
    dump_default_header(fd);
    if(content_type[0])
        fprintf(fd,"Content-type: %s%c",content_type,LF);
    if(last_modified[0])
        fprintf(fd,"Last-modified: %s GMT%c",last_modified,LF);
    if(content_length >= 0) 
        fprintf(fd,"Content-length: %d%c",content_length,LF);
    if(location[0])
        fprintf(fd,"Location: %s%c",location,LF);

    fprintf(fd,"%c",LF);
}
